@@ -30,22 +30,37 @@ def printf(fmt, *args, file=sys.stdout):
 
 def is_html(response):
     """ Return True if the response is a HTML webpage """
-    return "text/html" in response.headers["Content-Type"]
+    return (
+        "Content-Type" in response.headers
+        and "text/html" in response.headers["Content-Type"]
+    )
+
+
+def is_safe_path(path):
+    """ Prevent directory traversal attacks """
+    if path.startswith("/"):
+        return False
+
+    safe_path = os.path.expanduser("~")
+    return (
+        os.path.commonpath(
+            (os.path.realpath(os.path.join(safe_path, path)), safe_path)
+        )
+        == safe_path
+    )
 
 
 def get_indexed_files(response):
     """ Return all the files in the directory index webpage """
     html = bs4.BeautifulSoup(response.text, "html.parser")
     files = []
-    bad = {".", "..", "../"}
 
     for link in html.find_all("a"):
         url = urllib.parse.urlparse(link.get("href"))
 
         if (
             url.path
-            and url.path not in bad
-            and not url.path.startswith("/")
+            and is_safe_path(url.path)
             and not url.scheme
             and not url.netloc
         ):
@@ -102,6 +117,8 @@ def get_referenced_sha1(obj_file):
             objs.append(item.sha.decode())
     elif isinstance(obj_file, dulwich.objects.Blob):
         pass
+    elif isinstance(obj_file, dulwich.objects.Tag):
+        pass
     else:
         printf(
             "error: unexpected object type: %r\n" % obj_file, file=sys.stderr
@@ -332,7 +349,7 @@ class FindRefsWorker(DownloadWorker):
             r"(refs(/[a-zA-Z0-9\-\.\_\*]+)+)", response.text
         ):
             ref = ref[0]
-            if not ref.endswith("*"):
+            if not ref.endswith("*") and is_safe_path(ref):
                 tasks.append(".git/%s" % ref)
                 tasks.append(".git/logs/%s" % ref)
 
@@ -412,7 +429,7 @@ def fetch_git(url, directory, jobs, retry, timeout, http_headers):
     if not valid:
         printf(error_message, url, "/.git/HEAD", file=sys.stderr)
         return 1
-    elif not re.match(r'^(ref:.*|[0-9a-f]{40}$)', response.text.strip()):
+    elif not re.match(r"^(ref:.*|[0-9a-f]{40}$)", response.text.strip()):
         printf(
             "error: %s/.git/HEAD is not a git HEAD file\n",
             url,
@@ -563,7 +580,7 @@ def fetch_git(url, directory, jobs, retry, timeout, http_headers):
     if os.path.exists(index_path):
         index = dulwich.index.Index(index_path)
 
-        for entry in index.iterblobs():
+        for entry in index.iterobjects():
             objs.add(entry[1].decode())
 
     # use packs to find more objects to fetch, and objects that are packed
